#include <algorithm>
#include "kktdirect.h"
#include "kktsupport.h"

void KKT_detect_constraints(KKT_INT n, 
                            const KKT_INT *colstart,
                            const KKT_INT *rowindex,
                            const double *value,
                            char *constrained)
{
   for(KKT_INT i=0; i<n; ++i){
      constrained[i]=1;
      for(KKT_INT a=colstart[i]; a<colstart[i+1]; ++a){
         if(rowindex[a]==i && (!value || value[a]>0)){
            constrained[i]=0;
            break;
         }
      }
   }
}

// calculates elimination tree given ordering and its inverse permutation.
// parent[] is the standard description (for each node, what is its parent in the tree, -1 if it is a root)
// children[i] is a list of the children of node i in the etree, *IN THE ORIGINAL ORDER OF ELIMINATION*.
static void KKT_etree(KKT_INT n, 
                      const KKT_INT *colstart,
                      const KKT_INT *rowindex,
                      const KKT_INT *ordering,
                      const KKT_INT *inv, // inverse of ordering
                      KKT_INT *parent,
                      std::vector<std::vector<KKT_INT> > &children)
{
   if(n<1) return;
   for(KKT_INT i=0; i<n; ++i) parent[i]=-1;
   std::vector<KKT_INT> ancestor(n,-1); // short-cut version of parent
   children.resize(0); children.resize(n);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=ordering[a]; // i is the a'th node in ordering
      assert(inv[i]==a);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b]; // j is a neighbour of i in matrix
         KKT_INT c=inv[j];  // i.e. j was the c'th node in ordering
         if(c<a){ // if j was ordered before i
            KKT_INT k=j;
            while(ancestor[k]!=-1){
               KKT_INT next_k=ancestor[k];
               if(next_k==i) goto skip_this_neighbour;
               assert(next_k!=k);
               ancestor[k]=i;
               k=next_k;
               assert(inv[k]<a);
            }
            assert(parent[k]==-1);
            parent[k]=i;
            ancestor[k]=i;
            children[i].push_back(inv[k]); // NOTE: we put k's rank in the ordering, not k itself (will fix this later)
         }
         skip_this_neighbour: ;
      }
   }
   for(KKT_INT i=0; i<n; ++i){
      sort(children[i].begin(), children[i].end()); // sort children by occurrence in ordering
      for(KKT_UINT a=0; a<children[i].size(); ++a){
         children[i][a]=ordering[children[i][a]]; // translate back to actual node labels
         assert(parent[children[i][a]]==i);
      }
   }
}

void KKT_postorder(KKT_INT n, 
                   const KKT_INT *colstart,
                   const KKT_INT *rowindex,
                   const KKT_INT *ordering,
                   KKT_INT *postordering,
                   KKT_INT *inverseordering,
                   KKT_INT *etreeparent)
{
   if(n<1) return;
   // inverse of initial ordering
   for(KKT_INT a=0; a<n; ++a){
      inverseordering[ordering[a]]=a;
   }
   // get the initial ordering's elimination tree (actually, a forest)
   std::vector<std::vector<KKT_INT> > children(n);
   KKT_etree(n, colstart, rowindex, ordering, inverseordering, etreeparent, children);
   // depth-first-search from each root (creating postordering in reverse)
   std::vector<KKT_INT> stack;
   stack.reserve(n);
   KKT_INT k=n; // where we write into postordering
   DEBUG(std::vector<bool> postordered(n,false));
   for(KKT_INT r=n-1; r>=0; --r){
      if(etreeparent[r]!=-1) continue; // just look for the roots of the elimination "tree" (really, forest)
      stack.push_back(r);
      while(!stack.empty()){
         KKT_INT i=stack.back(); stack.pop_back();
         assert(!postordered[i]);
         assert(k>0);
         --k; postordering[k]=i;
         DEBUG(postordered[i]=true);
         for(KKT_INT b=children[i].size()-1; b>=0; --b){
            stack.push_back(children[i][b]);
         }
      }
   }
   // now update inverse_ordering to match postordering (note that etree_parent is unchanged)
   for(KKT_INT a=0; a<n; ++a){
      inverseordering[postordering[a]]=a;
   }
}

void KKT_symbolically_factor_uplooking(KKT_INT n,
                                       const KKT_INT *colstart,
                                       const KKT_INT *rowindex,
                                       const KKT_INT *postordering,
                                       const KKT_INT *inverseordering,
                                       const KKT_INT *etreeparent,
                                       KKT_INT *factorcolstart)
{
   if(n<1) return;
   // not the fastest approach, but simplest to code up: compute each row's structure, accumulate column counts as we go
   std::fill_n(factorcolstart, n+1, 0); // for now, store column length in factorcolstart rather than start of column
   Marker rowmark(n);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      // calculate structure of L(i,:)
      rowmark.clear();
      rowmark.set(a);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         KKT_INT c=inverseordering[j];
         if(c>=a) continue;
         while(!rowmark[c]){
            rowmark.set(c);
            ++factorcolstart[c+1];
            j=etreeparent[j];
            c=inverseordering[j];
         }
      }
   }
   // do the cumulative sum to get where each column of the factor actually starts
   for(KKT_INT a=2; a<n+1; ++a){
      factorcolstart[a]+=factorcolstart[a-1];
   }
}

void KKT_symbolically_factor_signed_cholesky(KKT_INT n,
                                             const KKT_INT *colstart,
                                             const KKT_INT *rowindex,
                                             const KKT_INT *postordering,
                                             const KKT_INT *inverseordering,
                                             const KKT_INT *etreeparent,
                                             KKT_INT *factorcolstart)
{
   if(n<1) return;
   // not the fastest approach, but simplest to code up: compute each row's structure, accumulate column counts as we go
   std::fill_n(factorcolstart, n+1, 0); // for now, store column length in factorcolstart rather than start of column
   Marker rowmark(n);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      // calculate structure of L(i,:)
      rowmark.clear();
      rowmark.set(a);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         KKT_INT c=inverseordering[j];
         if(c>=a) continue;
         while(!rowmark[c]){
            rowmark.set(c);
            ++factorcolstart[c+1];
            j=etreeparent[j];
            c=inverseordering[j];
         }
      }
   }
   // do the cumulative sum to get where each column of the factor actually starts
   // also, add 1 to account for extra diagonal entry stored in signed Cholesky factorization
   for(KKT_INT a=1; a<n+1; ++a){
      factorcolstart[a]+=factorcolstart[a-1]+1;
   }
}

void KKT_symbolically_factor_supernodal_initial(KKT_INT n,
                                                const KKT_INT *colstart,
                                                const KKT_INT *rowindex,
                                                const KKT_INT *postordering,
                                                const KKT_INT *inverseordering,
                                                const KKT_INT *etreeparent,
                                                KKT_INT *numsupernodes,
                                                KKT_INT *supernodestart,
                                                KKT_INT *supernodemap,
                                                KKT_INT *supercolstart,
                                                KKT_INT *factorcolstart,
                                                KKT_INT *diagblockstart)
{
   if(n<1) return;
   // idea: begin with all nodes in one super node.
   //       break apart nodes that aren't along a path in the etree
   //       calculate structure of L row by row, breaking up supernodes where there's a difference in structure
   //       and accumulate column counts as we go (for final processing into supercolstart and factorcolstart)
   std::vector<KKT_INT> columncount(n, 0);
   std::vector<bool> supernodebreak(n, false); // supernodebreak[a] indicates if node a+1 is in a different supernode than node a
                                               // (under given postordering). We begin with all nodes in one supernode.
   supernodebreak[n-1]=true; // special end case (sentinel value)
   Marker rowmark(n+1);
   std::vector<KKT_INT> rowlist;
   rowlist.reserve(n);
   for(KKT_INT a=0; a<n; ++a){
      // break a from a+1 if not directly related in etree
      if(a+1<n && etreeparent[postordering[a]]!=postordering[a+1]){
         supernodebreak[a]=true;
      }
      // calculate structure of L(a,:)
      KKT_INT i=postordering[a];
      rowmark.clear();
      rowmark.set(a);
      rowlist.resize(0);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         if(inverseordering[j]>=a) continue;
         KKT_INT k=j;
         for(;;){
            KKT_INT c=inverseordering[k];
            if(rowmark[c]) break;
            rowmark.set(c);
            rowlist.push_back(c);
            k=etreeparent[k];
         }
      }
      // now zip through those nonzeros breaking apart supernodes where necessary and accumulating column counts
      for(KKT_UINT p=0; p<rowlist.size(); ++p){
         KKT_INT b=rowlist[p];
         assert(b<a);
         assert(rowmark[b]);
         ++columncount[b];
         if(b>0 && !rowmark[b-1]) supernodebreak[b-1]=true;
         if(!rowmark[b+1]) supernodebreak[b]=true;
      }
   }
   // tally up the supernodes and decide supernodestart and supernodemap
   *numsupernodes=0;
   supernodestart[0]=0;
   for(KKT_INT a=0; a<n; ++a){
      supernodemap[a]=*numsupernodes;
      if(supernodebreak[a]){
         ++(*numsupernodes);
         supernodestart[*numsupernodes]=a+1;
      }
   }

   // sanity check: make sure each supernode is a path in the etree (I think it should be)
   for(KKT_INT s=0; s<*numsupernodes; ++s){
      if(supernodestart[s+1]>supernodestart[s]+1){
         for(KKT_INT a=supernodestart[s]; a<supernodestart[s+1]-1; ++a){
            assert(etreeparent[postordering[a]]==postordering[a+1]);
         }
      }
   }

   // figure out supercolstart (where rowindices for supercolumns are stored)
   supercolstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      supercolstart[a+1]=supercolstart[a]+columncount[supernodestart[a+1]-1]; // don't include nonzeros inside diagonal block
   }
   // figure out factorcolstart (where values for factor are stored)
   factorcolstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      factorcolstart[a+1]=factorcolstart[a]+(supernodestart[a+1]-supernodestart[a])*columncount[supernodestart[a+1]-1];
   }
   // and finally diagonal blocks
   diagblockstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      diagblockstart[a+1]=diagblockstart[a]+sqr(supernodestart[a+1]-supernodestart[a]);
   }
}

void KKT_symbolically_factor_supernodal_initial_for_signed_cholesky(KKT_INT n,
                                                                    const KKT_INT *colstart,
                                                                    const KKT_INT *rowindex,
                                                                    KKT_INT max_supernode_size,
                                                                    KKT_INT *postordering,
                                                                    KKT_INT *inverseordering,
                                                                    KKT_INT *etreeparent,
                                                                    const char *constrained,
                                                                    KKT_INT *numsupernodes,
                                                                    KKT_INT *supernodestart,
                                                                    KKT_INT *supernodemap,
                                                                    KKT_INT *supercolstart,
                                                                    KKT_INT *factorcolstart,
                                                                    KKT_INT *diagblockstart)
{
   if(n<1) return;
   // idea: begin with all nodes in one super node.
   //       break apart nodes that aren't along a path in the etree
   //       calculate structure of L row by row, breaking up supernodes where there's a difference in structure
   //       and accumulate column counts as we go (for final processing into supercolstart and factorcolstart)
   std::vector<KKT_INT> columncount(n, 0);
   std::vector<bool> supernodebreak(n, false); // supernodebreak[a] indicates if node a+1 is in a different supernode than node a
                                               // (under given postordering). We begin with all nodes in one supernode.
   supernodebreak[n-1]=true; // special end case (sentinel value)
   Marker rowmark(n+1);
   std::vector<KKT_INT> rowlist;
   rowlist.reserve(n);
   for(KKT_INT a=0; a<n; ++a){
      // break a from a+1 if not directly related in etree
      if(a+1<n && etreeparent[postordering[a]]!=postordering[a+1]){
         supernodebreak[a]=true;
      }
      // calculate structure of L(a,:)
      KKT_INT i=postordering[a];
      rowmark.clear();
      rowmark.set(a);
      rowlist.resize(0);
      for(KKT_INT b=colstart[i]; b<colstart[i+1]; ++b){
         KKT_INT j=rowindex[b];
         if(inverseordering[j]>=a) continue;
         KKT_INT k=j;
         for(;;){
            KKT_INT c=inverseordering[k];
            if(rowmark[c]) break;
            rowmark.set(c);
            rowlist.push_back(c);
            k=etreeparent[k];
         }
      }
      // now zip through those nonzeros breaking apart supernodes where necessary and accumulating column counts
      for(KKT_UINT p=0; p<rowlist.size(); ++p){
         KKT_INT b=rowlist[p];
         assert(b<a);
         assert(rowmark[b]);
         ++columncount[b];
         if(b>0 && !rowmark[b-1]) supernodebreak[b-1]=true;
         if(!rowmark[b+1]) supernodebreak[b]=true;
      }
   }
   // check within each supernode for constrained versus unconstrained nodes, determining reordering
   std::vector<KKT_INT> local_unconstrained, local_constrained;
   for(KKT_INT a=0; a<n; ){
      //std::cout<<"looking at supernode starting at "<<a<<" (ordering=="<<postordering[a]<<")"<<std::endl;
      local_unconstrained.resize(0);
      local_constrained.resize(0);
      if(constrained[postordering[a]]){
         local_constrained.push_back(postordering[a]);
      }else{
         local_unconstrained.push_back(postordering[a]);
      }
      // look through this supernode, splitting it up into constrained and unconstrained nodes
      KKT_INT b=a;
      for(; !supernodebreak[b]; ++b){
         if(constrained[postordering[b+1]]){
            local_constrained.push_back(postordering[b+1]);
         }else{
            local_unconstrained.push_back(postordering[b+1]);
         }
      }
      // if we have a mix inside this supernode...
      if(!local_unconstrained.empty() && !local_constrained.empty()){
         // rewrite this section of postordering so that all the unconstrained nodes come first
         KKT_INT c=a-1;
         for(unsigned int p=0; p<local_unconstrained.size(); ++p){
            postordering[c+1]=local_unconstrained[p];
            ++c;
         }
         supernodebreak[c]=true; // split the supernode here, between unconstrained and constrained
         for(unsigned int p=0; p<local_constrained.size(); ++p){
            postordering[c+1]=local_constrained[p];
            ++c;
         }
         assert(c==b);
      }
      // and on to the next original supernode
      a=b+1;
   }
   // tally up the supernodes and decide supernodestart (and break up supernodes that are too large as we go)
   *numsupernodes=0;
   supernodestart[0]=0;
   for(KKT_INT a=0; a<n; ++a){
      supernodemap[a]=*numsupernodes;
      if(supernodebreak[a] || a-supernodestart[*numsupernodes]>=max_supernode_size){
         ++(*numsupernodes);
         supernodestart[*numsupernodes]=a+1;
      }
   }
   // figure out supercolstart (where rowindices for supercolumns are stored)
   supercolstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      supercolstart[a+1]=supercolstart[a]+columncount[supernodestart[a+1]-1]; // don't include nonzeros inside diagonal block
   }
   // figure out factorcolstart (where values for factor are stored)
   factorcolstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      factorcolstart[a+1]=factorcolstart[a]+(supernodestart[a+1]-supernodestart[a])*columncount[supernodestart[a+1]-1];
   }
   // and diagonal blocks
   diagblockstart[0]=0;
   for(KKT_INT a=0; a<*numsupernodes; ++a){
      diagblockstart[a+1]=diagblockstart[a]+sqr(supernodestart[a+1]-supernodestart[a]);
   }
   // last of all, fix up inverseordering and elimination tree (which may have been modified by reordering)
   // There's undoubtedly a smart way to update, but for now, just recompute from scratch
   for(KKT_INT a=0; a<n; ++a){
      inverseordering[postordering[a]]=a;
   }
   std::vector<std::vector<KKT_INT> > children(n);
   KKT_etree(n, colstart, rowindex, postordering, inverseordering, etreeparent, children);
}

void KKT_symbolically_factor_supernodal_final(KKT_INT n,
                                              const KKT_INT *colstart,
                                              const KKT_INT *rowindex,
                                              const KKT_INT *postordering,
                                              const KKT_INT *inverseordering,
                                              const KKT_INT *etreeparent,
                                              KKT_INT numsupernodes,
                                              const KKT_INT *supernodestart,
                                              const KKT_INT *supernodemap,
                                              const KKT_INT *supercolstart,
                                              KKT_INT *superrowindex)
{
   if(n<1) return;
   std::vector<KKT_INT> supercolend(numsupernodes);
   std::copy(supercolstart, supercolstart+numsupernodes, supercolend.begin());
   Marker marker(n), supermarker(numsupernodes);
   for(KKT_INT a=0; a<n; ++a){
      KKT_INT i=postordering[a];
      marker.clear();
      marker.set(a);
      supermarker.clear();
      supermarker.set(supernodemap[a]); // don't include entries in the diagonal block
      for(KKT_INT p=colstart[i]; p<colstart[i+1]; ++p){
         KKT_INT j=rowindex[p];
         KKT_INT b=inverseordering[j];
         if(b>=a) continue;
         KKT_INT s=supernodemap[b];
         while(!marker[b]){
            marker.set(b);
            if(!supermarker[s]){
               KKT_INT q=supercolend[s];
               ++supercolend[s];
               superrowindex[q]=a;
               supermarker.set(s);
            }
            j=etreeparent[j];
            b=inverseordering[j];
            s=supernodemap[b];
         }
      }
   }
   DEBUG(
   for(KKT_INT s=0; s<numsupernodes; ++s){
      assert(supercolend[s]==supercolstart[s+1]);
   })
}

void KKT_bogus_supernodal_initial(KKT_INT n,
                                  const KKT_INT *colstart,
                                  const KKT_INT *rowindex,
                                  const KKT_INT *postordering,
                                  const KKT_INT *inverseordering,
                                  const KKT_INT *etreeparent,
                                  KKT_INT *numsupernodes,
                                  KKT_INT *supernodestart,
                                  KKT_INT *supernodemap,
                                  KKT_INT *supercolstart,
                                  KKT_INT *factorcolstart,
                                  KKT_INT *diagblockstart)
{
   KKT_symbolically_factor_uplooking(n, colstart, rowindex, postordering, inverseordering, etreeparent, factorcolstart);
   *numsupernodes=n;
   for(KKT_INT i=0; i<n; ++i){
      supernodemap[i]=i;
   }
   for(KKT_INT i=0; i<n+1; ++i){
      supernodestart[i]=i;
      supercolstart[i]=factorcolstart[i];
      diagblockstart[i]=i;
   }
}

